#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import logging


__author__ = 'EvgenySafronov <division494@gmail.com>'

log = logging.getLogger(__name__)


BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


COLORS = {
    'DEBUG': WHITE,
    'INFO': GREEN,
    'WARNING': YELLOW,
    'CRITICAL': YELLOW,
    'ERROR': RED
}


class ColoredFormatter(logging.Formatter):
    def __init__(self, msg, colored=True):
        logging.Formatter.__init__(self, msg)
        self.colored = colored

    def format(self, record):
        levelname = record.levelname
        if self.colored and levelname in COLORS:
            record.msg = COLOR_SEQ % (30 + COLORS[levelname]) + str(record.msg) + RESET_SEQ
        return logging.Formatter.format(self, record)


def interactiveEmit(self, record):  # pragma: no cover
    # Monkey patch Emit function to avoid new lines between records
    try:
        if str(record.msg).endswith('... '):
            fs = '%s'
        else:
            fs = '%s\n'
        msg = self.format(record)
        stream = self.stream
        if not hasattr(logging, '_unicode') or not logging._unicode:  # if no unicode support...
            stream.write(fs % msg)
        else:
            try:
                if isinstance(msg, unicode) and getattr(stream, 'encoding', None):
                    ufs = fs.decode(stream.encoding)
                    try:
                        stream.write(ufs % msg)
                    except UnicodeEncodeError:
                        stream.write((ufs % msg).encode(stream.encoding))
                else:
                    stream.write(fs % msg)
            except UnicodeError:
                stream.write(fs % msg.encode("UTF-8"))
        self.flush()
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:
        self.handleError(record)



try:
    import httplib
except ImportError:
    import http.client as httplib  # pylint: disable=F0401

import msgpack

from tornado import gen
from tornado import httputil
from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPError
from tornado.httpclient import HTTPRequest

from cocaine.proxy.helpers import extract_app_and_event
from cocaine.proxy.helpers import fill_response_in
from cocaine.proxy.helpers import pack_httprequest

from cocaine.proxy.plugin import IPlugin
from cocaine.proxy.plugin import PluginApplicationError
from cocaine.proxy.plugin import PluginConfigurationError
from cocaine.proxy.plugin import PluginNoSuchApplication


def is_mds_stid(stid):
    _, _, tail = stid.split(".", 2)
    return tail.startswith('E') and ':' in tail


class MDSExec(IPlugin):
    def __init__(self, proxy, config):
        super(MDSExec, self).__init__(proxy)
        try:
            self.srw_host = config["srw_host"]
            self.filter_mds_stid = config.get("filter_stid", True)
            self.srw_httpclient = AsyncHTTPClient()
        except KeyError as err:
            raise PluginConfigurationError(self.name(), "option required %s" % err)

    @staticmethod
    def name():
        return "mds"

    def match(self, request):
        if "X-Srw-Key" in request.headers and "X-Srw-Key-Type" in request.headers and "X-Srw-Namespace" in request.headers:
            return not self.filter_mds_stid or is_mds_stid(request.headers["X-Srw-Key"])
        return False

    @gen.coroutine
    def process(self, request):
        name, event = extract_app_and_event(request)
        timeout = self.proxy.get_timeout(name, event)
        # as MDS proxy bypasses the mechanism of routing groups
        # the proxy is responsible to provide this feature
        name = self.proxy.resolve_group_to_version(name)
        headers = request.headers
        namespace = headers["X-Srw-Namespace"]
        key = headers["X-Srw-Key"]

        mds_request_headers = httputil.HTTPHeaders()
        if "Authorization" in request.headers:
            mds_request_headers["Authorization"] = request.headers["Authorization"]

        traceid = getattr(request, "traceid", None)
        if traceid is not None:
            mds_request_headers["X-Request-Id"] = traceid

        srw_request = HTTPRequest("%s/exec-%s/%s/%s/stid/%s?timeout=%d" % (self.srw_host, namespace, name, event, key, timeout),
                                  method="POST",
                                  headers=mds_request_headers,
                                  body=msgpack.packb(pack_httprequest(request)),
                                  allow_ipv6=True,
                                  request_timeout=timeout)

        try:
            # NOTE: we can do it in a streaming way
            resp = yield self.srw_httpclient.fetch(srw_request)
            code, reply_headers, body = decode_chunked_encoded_reply(resp)
            fill_response_in(request, code,
                             httplib.responses.get(code, httplib.OK),
                             body, reply_headers)
        except HTTPError as err:
            if err.code == 404:
                raise PluginNoSuchApplication("worker was not found")

            if err.code == 500:
                raise PluginApplicationError(42, 42, "worker replied with error")

            if err.code == 401:
                fill_response_in(request, err.code,
                                 httplib.responses.get(err.code, httplib.OK),
                                 err.response.body, err.response.headers)
                return

            raise err


def decode_chunked_encoded_reply(resp):
    # read_size is set to to prevent overead from BytesIO
    # in this case the rest of the buffer is not packed data
    code, raw_headers = msgpack.Unpacker(resp.buffer, read_size=1).unpack()
    body = resp.buffer.read(None)
    headers = httputil.HTTPHeaders(raw_headers)
    return code, headers, body



#!/usr/bin/env python
#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import logging
import errno

from cocaine.tools.dispatch import cli


__author__ = 'EvgenySafronov <division494@gmail.com>'


__doc__ = '''Provides helpful tools for management, viewing, uploading and other actions with
cocaine applications and services.'''


log = logging.getLogger('cocaine.tools')


def main():
    try:
        cli()
    except KeyboardInterrupt:
        log.error('Terminated by user')
        exit(errno.EINTR)
    except Exception as err:
        log.exception('Unknown error occurred - %s', err)
        exit(128)


if __name__ == '__main__':
    main()



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division

from collections import defaultdict
import datetime
import json
import fnmatch
import os
import time
import socket

from tornado import gen

from cocaine.decorators import coroutine
from cocaine.tools.error import ToolsError
from . import mql

__author__ = 'Evgeny Safronov <division494@gmail.com>'


def split_by_groups(items, split_by=10):
    return (items[i - split_by:i] for i in xrange(split_by, len(items) + split_by, split_by))


class Node(object):
    def __init__(self, node=None):
        self.node = node

    @coroutine
    def execute(self):
        raise NotImplementedError()


class Locate(object):
    def __init__(self, locator, name):
        self.locator = locator
        if not name:
            raise ValueError("option `name` must be specified")
        self.name = name

    @coroutine
    def execute(self):
        channel = yield self.locator.resolve(self.name)
        endpoints, version, api = yield channel.rx.get()
        result = {
            "endpoints": ["%s:%d" % (addr, port) for addr, port in endpoints],
            "version": version,
            "api": dict((num, method[0]) for num, method in api.items())
        }
        raise gen.Return(result)


class Cluster(object):
    def __init__(self, locator, resolve=True):
        self.locator = locator
        self.resolve = resolve

    @coroutine
    def execute(self):
        channel = yield self.locator.cluster()
        result = yield channel.rx.get()
        if self.resolve:
            raise gen.Return(result)

        converted_result = {}
        for uuid, (addr, port) in result.items():
            try:
                host = socket.gethostbyaddr(addr)[0]
                converted_result[uuid] = [host, port]
            except (socket.gaierror, IOError):
                converted_result[uuid] = [addr, port]

        raise gen.Return(converted_result)


class Routing(object):
    extent = pow(2, 32)

    def __init__(self, locator, name=None):
        self.locator = locator
        self.name = name

    def generate_group(self, body):
        if len(body) == 0:
            return {}
        apps = defaultdict(int)
        # initialize with maximum value
        # from the routing
        prev = body[-1][0] - Routing.extent
        for value, app in body:
            apps[app] += (value - prev)
            prev = value

        output = dict((a, w / Routing.extent) for a, w in apps.items())
        return output

    @coroutine
    def execute(self):
        uid = "tools:%s_%d_%f" % (socket.gethostname(), os.getpid(), time.time())
        channel = yield self.locator.routing(uid, True)
        rings = yield channel.rx.get()
        groups = {}
        if not self.name:
            for name, ring in rings.items():
                groups[name] = self.generate_group(ring)
        elif self.name in rings:
            groups[self.name] = self.generate_group(rings[self.name])
        else:
            raise ToolsError("No such group `%s` in the routing. "
                             "Probably you should refresh the locator" % self.name)

        raise gen.Return(groups)


class RuntimeMetrics(object):
    def __init__(self, ty, query, query_type, metrics):
        self._ty = ty
        if query is not None:
            if query_type == 'ast':
                self._query = json.loads(query)
            else:
                self._query = mql.compile_query(query)
        else:
            self._query = None
        self._metrics = metrics

    @coroutine
    def execute(self):
        channel = yield self._metrics.fetch(self._ty, self._query)
        metrics = yield channel.rx.get()
        raise gen.Return(metrics)


class NodeInfo(Node):
    def __init__(self, node, locator,
                 name=None, flags=0x1,
                 use_wildcard=False,
                 timeout=5):
        super(NodeInfo, self).__init__(node)
        self.locator = locator
        self._name = name
        self._flags = flags
        self._use_wildcard = use_wildcard
        self.timeout = datetime.timedelta(seconds=timeout)
        self.on_timeout_reply = {
            "error": "info was timed out",
            "state": "unresponsive",
            "meta": "the error was generated by tools"
        }

    @coroutine
    def execute(self):
        # name is provided and wildcard is switched off
        # so we use exact match
        if self._name and not self._use_wildcard:
            apps = [self._name]
        else:
            channel = yield self.node.list()
            apps = yield channel.rx.get()
            # wildcard has been already checked
            if self._name:
                apps = fnmatch.filter(apps, self._name)

        result = yield self.info(apps)
        raise gen.Return(result)

    @coroutine
    def info(self, apps):
        infos = {}
        for names in split_by_groups(apps, 25):
            futures = {}
            for app in names:
                res_future = (yield self.node.info(app, self._flags)).rx.get()
                futures[app] = gen.with_timeout(self.timeout, res_future)

            wait_iterator = gen.WaitIterator(**futures)
            while not wait_iterator.done():
                try:
                    info = yield wait_iterator.next()
                    infos[wait_iterator.current_index] = info
                except gen.TimeoutError:
                    infos[wait_iterator.current_index] = self.on_timeout_reply
                except Exception as err:
                    infos[wait_iterator.current_index] = str(err)

        result = {
            'apps': infos,
            'count': len(infos),
        }
        raise gen.Return(result)



#!/usr/bin/env python
#
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

__author__ = 'EvgenySafronov <division494@gmail.com>'


def getOption(name, default):
    value = default
    if name in sys.argv:
        index = sys.argv.index(name)
        if index < len(sys.argv) - 1:
            value = sys.argv[index + 1]
            if value == '=' and index + 1 < len(sys.argv) - 2:
                value = sys.argv[index + 2]
    elif name + '=' in sys.argv:
        index = sys.argv.index(name + '=')
        if index < len(sys.argv) - 1:
            value = sys.argv[index + 1]
    return value


if __name__ == '__main__':
    try:
        import sys
        import os
        from cocaine.services import Service
        from cocaine.asio import engine

        ADEQUATE_TIMEOUT = 0.25

        locateItems = {
            'app': ['manifests', ('app', )],
            'profile': ['profiles', ('profile',)],
            'runlist': ['runlists', ('runlist',)],
            'group': ['groups', ('group',)],
        }

        config = {
            'locateItem': getOption('--locator_type', 'app'),
            'host': getOption('--host', 'localhost'),
            'port': getOption('--port', '10053')
        }

        @engine.asynchronous
        def locateApps():
            apps = yield storage.find(*locateItems.get(config['locateItem']))
            with open('/tmp/1.txt', 'w') as fh:
                fh.write(' '.join(apps))
            if apps:
                print(' '.join(apps))

        storage = Service('storage', endpoints=[(config['host'], int(config['port']))])
        locateApps().get(timeout=ADEQUATE_TIMEOUT)
    except Exception as err:
        # Hidden log feature :)
        with open(os.devnull, 'w') as fh:
            fh.write(str(err))



from tornado import gen


class PluginException(Exception):
    pass


class PluginConfigurationError(PluginException):
    def __init__(self, name, message):
        super(PluginConfigurationError, self).__init__("configuration error for %s: %s" % (name, message))
        self.name = name
        self.message = message


class PluginNoSuchApplication(PluginException):
    pass


class PluginApplicationError(PluginException):
    def __init__(self, category, code, message):
        super(PluginApplicationError, self).__init__("[%d %d] %s" % (category, code, message))
        self.category = category
        self.code = code
        self.message = message


class IPlugin(object):
    @staticmethod
    def name():
        raise NotImplementedError()

    def __init__(self, proxy):
        self.proxy = proxy

    def match(self, request):
        raise NotImplementedError()

    @gen.coroutine
    def process(self, request):
        raise NotImplementedError()



import json

import click
import msgpack
from tornado import gen

from cocaine.decorators import coroutine
from cocaine.exceptions import ServiceError
from cocaine.tools.actions import Action

_COLLECTION_TVM = 'tvm'

_KEY_PUBLIC_KEYS = 'public_keys'


class View(Action):
    def __init__(self, storage):
        self._storage = storage

    @coroutine
    def execute(self):
        channel = yield self._storage.read(_COLLECTION_TVM, _KEY_PUBLIC_KEYS)
        keys = yield channel.rx.get()
        keys = msgpack.loads(keys)
        raise gen.Return(keys)


class Update(Action):
    def __init__(self, cid, tvm):
        self._cid = cid
        self._tvm = tvm

    @coroutine
    def execute(self):
        channel = yield self._tvm.download_keys(self._cid)
        keys = yield channel.rx.get()
        raise gen.Return(keys)


class Remove(Action):
    def __init__(self, key, storage):
        self._key = key
        self._storage = storage

    @coroutine
    def execute(self):
        if self._key is None:
            keys = []
        else:
            keys = yield View(self._storage).execute()
            try:
                keys.remove(self._key)
            except ValueError:
                pass
        keys = msgpack.dumps(keys)

        channel = yield self._storage.write(_COLLECTION_TVM, _KEY_PUBLIC_KEYS, keys, [])
        yield channel.rx.get()


class Edit(Action):
    def __init__(self, storage):
        self._storage = storage

    @coroutine
    def execute(self):
        try:
            content = yield View(self._storage).execute()
        except ServiceError:
            content = []

        updated = click.edit(json.dumps(content, indent=4))
        if updated is not None:
            updated = json.loads(updated)
            updated = msgpack.dumps(updated)
            channel = yield self._storage.write(_COLLECTION_TVM, _KEY_PUBLIC_KEYS, updated, [])
            yield channel.rx.get()


class Refresh(Action):
    def __init__(self, tvm):
        self._tvm = tvm

    @coroutine
    def execute(self):
        channel = yield self._tvm.refresh_keyring()
        yield channel.rx.get()



#
# Copyright (c) 2016+ Anton Matveenko <antmat@me.com>
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from collections import defaultdict
from datetime import datetime
import json

from tornado import gen

from cocaine.tools.error import ToolsError


class LoggingConfigurator(object):
    def __init__(self, logging_service):
        self.logging_service = logging_service

    @gen.coroutine
    def execute(self):
        raise NotImplementedError


class LoggingConfigListLoggers(LoggingConfigurator):
    @gen.coroutine
    def execute(self):
        channel = yield self.logging_service.list_loggers()
        filters = yield channel.rx.get()
        raise gen.Return({"loggers": filters})


class LoggingConfigSetFilter(LoggingConfigurator):
    def __init__(self, logging_service, logger_name, filter_def, ttl):
        super(LoggingConfigSetFilter, self).__init__(logging_service)
        if not logger_name:
            raise ToolsError("Logger name is required")
        if not filter_def:
            raise ToolsError("filter definition is required")
        if not ttl:
            raise ToolsError("filter TTL is required")
        self.name = logger_name
        try:
            self.filter_def = json.loads(filter_def)
        except:
            raise ToolsError("Filter definition is not parsable. It should be JSON array with at least one element")
        try:
            self.ttl = int(ttl)
        except:
            raise ToolsError("TTL should be numeric")
        if not isinstance(self.filter_def, (tuple, list)) or len(self.filter_def) == 0:
            raise ToolsError("Filter definition should be array and contain at least one element")

    @gen.coroutine
    def execute(self):
        channel = yield self.logging_service.set_filter(self.name, self.filter_def, self.ttl)
        filters = yield channel.rx.get()
        raise gen.Return(filters)


class LoggingConfigRemoveFilter(LoggingConfigurator):
    def __init__(self, logging_service, filter_id):
        super(LoggingConfigRemoveFilter, self).__init__(logging_service)
        if not filter_id:
            raise ToolsError("Filter id is required")
        self.filter_id = int(filter_id)

    @gen.coroutine
    def execute(self):
        channel = yield self.logging_service.remove_filter(self.filter_id)
        filters = yield channel.rx.get()
        raise gen.Return(filters)


class LoggingConfigListFilters(LoggingConfigurator):
    @gen.coroutine
    def execute(self):
        channel = yield self.logging_service.list_filters()
        filters = yield channel.rx.get()
        ret = defaultdict(list)
        for logger_name, filter_def, filter_id, deadline, disposition_id in filters:
            if disposition_id == 0:
                disposition = "local"
            elif disposition_id == 1:
                disposition = "clusterwide"
            else:
                disposition = "unknown disposition_id: {}".format(disposition_id)
            time = datetime.fromtimestamp(min(deadline, 2 ** 32))
            ret[logger_name].append({"id": filter_id, "deadline": str(time),
                                     "filter_definition": filter_def, "disposition": disposition})
        raise gen.Return(ret)


class LoggingConfigSetClusterFilter(LoggingConfigSetFilter):
    def __init__(self, logging_service, logger_name, filter_def, ttl):
        super(LoggingConfigSetClusterFilter, self).__init__(logging_service, logger_name, filter_def, ttl)

    @gen.coroutine
    def execute(self):
        channel = yield self.logging_service.set_cluster_filter(self.name, self.filter_def, self.ttl)
        filters = yield channel.rx.get()
        raise gen.Return(filters)



#!/usr/bin/env python
# encoding: utf-8
#
#    Copyright (c) 2011-2012 Andrey Sibiryov <me@kobology.ru>
#    Copyright (c) 2011-2015 Anton Tyurin <noxiouz@yandex.ru>
#    Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
#    Copyright (c) 2011-2013 Other contributors as noted in the AUTHORS file.
#
#    This file is part of Cocaine.
#
#    Cocaine is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Cocaine is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import re

from setuptools import setup


if 'DEB_BUILD_GNU_TYPE' in os.environ:
    tools_data = [
        ('/etc/bash_completion.d/', ["scripts/bash_completion.d/cocaine-tool"])
    ]
else:
    tools_data = []


try:
    with open('cocaine/tools/version.py', 'r') as version_file:
        version = re.search("__version__ = \"(.+)\"", version_file.read()).groups()[0]
except Exception as err:
    raise Exception("unable to identify version: %s" % err)


setup(
    name="cocaine-tools",
    version=version,
    author="Anton Tyurin",
    author_email="noxiouz@yandex.ru",
    maintainer='Evgeny Safronov',
    maintainer_email='division494@gmail.com',
    url="https://github.com/cocaine/cocaine-tools",
    description="Cocaine Tools for Cocaine Application Cloud.",
    long_description="Tools for deploying and managing applications in the cloud",
    license="LGPLv3+",
    platforms=["Linux", "BSD", "MacOS"],
    namespace_packages=['cocaine'],
    include_package_data=True,
    zip_safe=False,
    packages=[
        "cocaine",
        "cocaine.proxy",
        "cocaine.tools",
        "cocaine.tools.actions",
        "cocaine.tools.actions.access",
        "cocaine.tools.actions.mql",
        "cocaine.tools.helpers",
        "cocaine.tools.interactive",
        "cocaine.tools.plugins",
        "cocaine.tools.plugins.secure",
    ],
    entry_points={
        'console_scripts': [
            'cocaine-tool = cocaine.tools.cocaine_tool:main',
            'cocaine-tornado-proxy = cocaine.proxy.proxy:main',
        ]},
    install_requires=open('./requirements.txt').read(),
    tests_require=open('./tests/requirements.txt').read(),
    test_suite='nose.collector',
    classifiers=[
        'Programming Language :: Python',
        # 'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3.2',
        # 'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        "Programming Language :: Python :: Implementation :: CPython",
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
    ],
)



#
#    Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
#    Copyright (c) 2011-2013 Other contributors as noted in the AUTHORS file.
#
#    This file is part of Cocaine.
#
#    Cocaine is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Cocaine is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import abc
import json
import os
import subprocess
import tempfile

from cocaine.decorators import coroutine
from cocaine.tools.actions import auth
from cocaine.tools.actions import group
from cocaine.tools.actions import profile
from cocaine.tools.actions import runlist
from cocaine.tools.helpers.editor import locate_editor
from cocaine.tools.printer import printer

__author__ = 'Evgeny Safronov <division494@gmail.com>'


class BaseEditor(object):
    __metaclass__ = abc.ABCMeta

    EDITORS = ['vim', 'emacs', 'nano']

    @abc.abstractmethod
    def view(self):
        pass

    @abc.abstractmethod
    def upload(self, data):
        pass

    @coroutine
    def execute(self):
        with printer('Loading "%s"', self.name):
            content = yield self.view()

        with printer('Editing "%s"', self.name):
            with tempfile.NamedTemporaryFile(delete=False) as fh:
                name = fh.name
                fh.write(json.dumps(content, indent=4))

            ec = None

            # locate default editor
            default_editor = locate_editor()
            if default_editor is not None:
                default_editor in self.EDITORS and self.EDITORS.remove(default_editor)
                self.EDITORS.insert(0, default_editor)

            for editor in self.EDITORS:
                try:
                    ec = subprocess.call([editor, name])
                    break
                except OSError:
                    continue

            if ec is None:
                raise ValueError('cannot open runlist for editing - any of {0} editors not found'.format(self.EDITORS))
            if ec != 0:
                raise ValueError('editing failed with exit code {0}'.format(ec))

        with open(name) as fh:
            yield self.upload(fh.read())
        # Remove temp file
        os.unlink(name)


class RunlistEditor(BaseEditor, runlist.Specific):
    def __init__(self, *args, **kwargs):
        super(RunlistEditor, self).__init__(*args, **kwargs)

    def view(self):
        return runlist.View(self.storage, self.name).execute()

    def upload(self, data):
        return runlist.Upload(self.storage, self.name, data).execute()


class ProfileEditor(BaseEditor, profile.Specific):
    def __init__(self, *args, **kwargs):
        super(ProfileEditor, self).__init__(*args, **kwargs)

    def view(self):
        return profile.View(self.storage, self.name).execute()

    def upload(self, data):
        return profile.Upload(self.storage, self.name, data).execute()


class GroupEditor(BaseEditor, group.Specific):
    def __init__(self, *args, **kwargs):
        super(GroupEditor, self).__init__(*args, **kwargs)

    def view(self):
        return group.View(self.storage, self.name).execute()

    def upload(self, data):
        return group.Create(self.storage, self.name, data).execute()


class Editor(BaseEditor):
    def __init__(self, storage, name):
        super(Editor, self).__init__()
        self.name = name
        self._storage = storage

    def view(self):
        return auth.View(self._storage, self.name).execute()

    def upload(self, data):
        return auth.Create(self._storage, self.name, data).execute()



import json

import click
from cerberus import Validator

from cocaine.tools.error import ToolsError
from tornado import gen

from cocaine.decorators import coroutine

from . import Action


COLLECTION_TOKENS = 'tokens'
COLLECTION_GROUPS = 'auth-groups'


class Auth(object):
    """Auth groups is a lie"""
    SCHEMA = {
        'token': {
            'type': 'string',
            'required': True,
            'regex': '^[0-9]+:.+$',
        },
        'members': {
            'type': 'list',
            'required': True,
            'schema': {
                'type': 'string',
            },
        },
    }

    def __init__(self, storage):
        self._storage = storage  # HINT: storage.write(namespace, key, value, tags)
        self._validator = Validator(self.SCHEMA)

    def _validate(self, group_definition):
        self._validator.validate(group_definition)
        if self._validator.errors:
            raise ValueError('failed to validate format: {}'.format(self._validator.errors))

    @coroutine
    def create_group(self, name, token, force=False):
        self._validate({'token': token, 'members': []})
        if not force:
            groups = yield self.list_groups()
            if name in groups:
                raise ToolsError('authorization group already exists')

        channel = yield self._storage.write(COLLECTION_GROUPS, name, token, ['auth'])
        yield channel.rx.get()

    @coroutine
    def view_group(self, name):
        channel = yield self._storage.read(COLLECTION_GROUPS, name)
        token = yield channel.rx.get()

        channel = yield self._storage.find(COLLECTION_TOKENS, ['auth', 'tokens', name])
        members = yield channel.rx.get()

        raise gen.Return({
            'token': token,
            'members': members,
        })

    @coroutine
    def list_groups(self):
        channel = yield self._storage.find(COLLECTION_GROUPS, ['auth'])
        value = yield channel.rx.get()
        raise gen.Return(value)

    @coroutine
    def remove_group(self, name, drop):
        channel = yield self._storage.remove(COLLECTION_GROUPS, name)
        yield channel.rx.get()

        if drop:
            channel = yield self._storage.find(COLLECTION_TOKENS, ['auth', 'tokens', name])
            members = yield channel.rx.get()

            for service_name in members:
                yield self.remove_member(name, service_name)

    @coroutine
    def add_member(self, name, service):
        channel = yield self._storage.read(COLLECTION_GROUPS, name)
        token = yield channel.rx.get()

        channel = yield self._storage.write(
            COLLECTION_TOKENS, service, token,
            ['auth', 'tokens', name],
        )
        yield channel.rx.get()

    @coroutine
    def remove_member(self, name, service):
        channel = yield self._storage.remove(COLLECTION_TOKENS, service)
        yield channel.rx.get()

    @coroutine
    def edit_group(self, name, updated):
        old = yield self.view_group(name)

        # change only parameters present in updated
        new = {}
        new.update(old)
        new.update(updated)

        self._validate(new)

        new['members'] = set(new['members'])
        old['members'] = set(old['members'])

        # Update token only if it was changed.
        if new['token'] != old['token']:
            yield self.create_group(name, new['token'], force=True)

            # update token for members already present in group
            for member in new['members'] & old['members']:
                yield self.add_member(name, member)

        # Remove excluded members while adding new ones.
        for member in new['members'] - old['members']:
            yield self.add_member(name, member)
        for member in old['members'] - new['members']:
            yield self.remove_member(name, member)


class AuthAction(Action):
    method = None

    def __init__(self, storage, *args, **kwargs):
        self.instance = self.method.im_class(storage)
        self.args = args
        self.kwargs = kwargs

    @coroutine
    def execute(self):
        value = yield self.method(self.instance, *self.args, **self.kwargs)
        raise gen.Return(value)


class List(AuthAction):
    method = Auth.list_groups


class Create(AuthAction):
    method = Auth.create_group


class View(AuthAction):
    method = Auth.view_group


class Remove(AuthAction):
    method = Auth.remove_group


class AddMember(AuthAction):
    method = Auth.add_member


class ExcludeMember(AuthAction):
    method = Auth.remove_member


class Edit(AuthAction):
    method = Auth.edit_group

    @coroutine
    def execute(self):
        content = yield self.instance.view_group(*self.args, **self.kwargs)
        updated = click.edit(json.dumps(content, indent=4))
        if updated is not None:
            updated = json.loads(updated)
            self.kwargs['updated'] = updated
            yield super(Edit, self).execute()



import logging
import socket

from tornado import gen

from cocaine.decorators import coroutine
from cocaine.tools.error import ToolsError

from . import SecurePlugin


log = logging.getLogger(__name__)


class TVM(SecurePlugin):
    def __init__(self, repo, client_id, client_secret):
        super(TVM, self).__init__(repo)
        self._client_id = client_id
        self._client_secret = client_secret

        addrinfo = socket.getaddrinfo(socket.gethostname(), None)
        if len(addrinfo) == 0:
            raise ToolsError('failed to determine local IP address')

        self._ip = addrinfo[0][4][0]
        self._tvm = repo.create_service('tvm')

    def ty(self):
        return 'TVM'

    @coroutine
    def fetch_token(self):
        grant_type = 'client_credentials'

        channel = yield self._tvm.ticket_full(self._client_id, self._client_secret, grant_type, {})
        ticket = yield channel.rx.get()
        log.debug('exchanged client secret with TVM ticket')
        raise gen.Return(self._make_header(ticket))

    def _make_header(self, ticket):
        return '{} {}'.format(self.ty(), ticket)



import collections
import hashlib
import json
from operator import xor
import re
import struct

from tornado import httputil

CRLF = '\r\n'

SIZE_OF_CHUNK_FMT = '{:X}' + CRLF
TERM_CHUNK = '0' + CRLF + CRLF

URL_REGEX = re.compile(r"/([^/]*)/([^/?]*)(.*)")

TcpEndpoint = collections.namedtuple('TcpEndpoint', ["host", "port"])


class ProxyInvalidRequest(Exception):
    pass


def load_srw_config(path):
    with open(path, mode='r') as f:
        config = json.load(f)
    return config


class Endpoints(object):
    unix_prefix = "unix://"
    tcp_prefix = "tcp://"

    def __init__(self, endpoints):
        self.unix = []
        self.tcp = []
        for i in endpoints:
            if i.startswith(Endpoints.unix_prefix):
                self.unix.append(i[len(Endpoints.unix_prefix):])
            elif i.startswith(Endpoints.tcp_prefix):
                raw = i[len(Endpoints.tcp_prefix):]
                delim_count = raw.count(":")
                if delim_count == 0:  # no port
                    raise ValueError("Endpoint has to containt host:port: %s" % i)
                elif delim_count == 1:  # ipv4 or hostname
                    host, _, port = raw.partition(":")
                    self.tcp.append(TcpEndpoint(host=host, port=int(port)))
                elif delim_count > 1:  # ipv6
                    host, _, port = raw.rpartition(":")
                    if host[0] != "[" or host[-1] != "]":
                        raise ValueError("Invalid IPv6 address %s" % i)
                    self.tcp.append(TcpEndpoint(host=host.strip("[]"), port=int(port)))
            else:
                raise ValueError("Endpoint has to begin either unix:// or tcp:// %s" % i)


def write_chunked(request, chunk):
    request.connection.write(SIZE_OF_CHUNK_FMT.format(len(chunk)))
    request.connection.write(chunk)
    request.connection.write(CRLF)


def finalize_response(request, code, status):
    request.connection.finish()
    request.logger.info("finish request: %d %s %.2fms",
                        code, status, 1000.0 * request.request_time())


def finalize_chunked_response(request, code, status):
    request.connection.write(TERM_CHUNK)
    finalize_response(request, code, status)


def fill_response_in(request, code, status, message, headers=None, chunked=False):
    headers = headers or httputil.HTTPHeaders()

    if not ("Content-Length" in headers or chunked):
        content_length = str(len(message))
        request.logger.debug("Content-Length header was generated by the proxy: %s", content_length)
        headers.add("Content-Length", content_length)

    headers.add("X-Powered-By", "Cocaine")
    headers["X-XSS-Protection"] = "1; mode=block"

    if not chunked:
        request.logger.debug("Content-Length: %s", headers["Content-Length"])

    if getattr(request, "traceid", None) is not None:
        headers.add("X-Request-Id", request.traceid)

    if request.method == "HEAD":
        message = None

    request.connection.write_headers(
        # start_line
        httputil.ResponseStartLine(request.version, code, status),
        # headers
        headers,
        # data
        message)

    if not chunked:
        finalize_response(request, code, status)


def pack_httprequest(request):
    headers = [(item.key, item.value) for item in request.cookies.itervalues()]
    headers.extend(request.headers.items())
    packed = request.method, request.uri, request.version.split("/")[1], headers, request.body
    return packed


def extract_app_and_event(request):
    if "X-Cocaine-Service" in request.headers and "X-Cocaine-Event" in request.headers:
        request.logger.debug('dispatch by headers')
        name = request.headers['X-Cocaine-Service']
        event = request.headers['X-Cocaine-Event']
        if name == '' or event == '':
            raise ProxyInvalidRequest("Invalid request: empty dispatch headers")
        return name, event

    # it's better to drop this way of dispatch
    # to prevent modification of request.path
    request.logger.debug('dispatch by uri')
    match = URL_REGEX.match(request.uri)
    if match is None:
        raise ProxyInvalidRequest("Invalid request")

    name, event, other = match.groups()
    if name == '' or event == '':
        raise ProxyInvalidRequest("Invalid request")

    # Drop from query appname and event's name
    if not other.startswith('/'):
        other = "/" + other
    request.uri = other
    request.path, _, _ = other.partition("?")
    # one more call of extract_app_and_event will not modify request.path
    request.headers['X-Cocaine-Service'] = name
    request.headers['X-Cocaine-Event'] = event
    return name, event


def parse_locators_endpoints(endpoint):
    host, _, port = endpoint.rpartition(":")
    if host and port:
        try:
            return (host, int(port))
        except ValueError:
            pass

    raise Exception("invalid endpoint: %s" % endpoint)


def upper_bound(l, value):
    lo = 0
    hi = len(l)
    while lo < hi:
        mid = (lo + hi) // 2
        if l[mid][0] < value:
            lo = mid + 1
        else:
            hi = mid
    return lo


def header_to_seed(value):
    m = hashlib.new("md5")
    m.update(value)
    # 4 bytes XOR
    return reduce(xor, struct.unpack("@IIII", m.digest()), 0)



import json
import logging
import re

log = logging.getLogger('cocaine.tools')

# Token types.
#
# EOF (end-of-file) token is used to indicate that there is no more input left for lexical
# analysis.
LITERAL, NUMBER, FUNCTION, AND, OR, EQ, GE, DOT, COMMA, LPAREN, RPAREN, EOF = (
    'LITERAL',
    'NUMBER',
    'FUNCTION',
    'AND',
    'OR',
    'EQ',
    'GE',
    'DOT',
    'COMMA',
    'LPAREN',
    'RPAREN',
    'EOF',
)

OPERATORS = {
    (COMMA, ','),
    (LPAREN, '('),
    (RPAREN, ')'),
    (AND, '&&'),
    (OR, '||'),
    (EQ, '=='),
    (GE, '>='),
    (FUNCTION, 'contains'),
    (FUNCTION, 'name'),
    (FUNCTION, 'type'),
    (FUNCTION, 'tag'),
}


class Token(object):
    def __init__(self, ty, value):
        self.type = ty
        self.value = value

    def __str__(self):
        """
        String representation of the class instance.

        Examples:
            Token(LITERAL, name)
            Token(LPAREN, '(')
            Token(RPAREN, ')')
        """
        return 'Token({type}, {value})'.format(type=self.type, value=repr(self.value))

    def __repr__(self):
        return 'Token(type: {type}, value: {value})'.format(type=self.type, value=repr(self.value))


class LiteralToken(Token):
    def __init__(self, value):
        try:
            super(LiteralToken, self).__init__(NUMBER, float(value))
        except ValueError:
            if not re.match(r'([\w.-_])+', value):
                raise SyntaxError('Invalid literal token')
            super(LiteralToken, self).__init__(LITERAL, value)


class AST(object):
    pass


class Const(AST):
    def __init__(self, token):
        self.token = token
        self.value = token.value

    def visit(self):
        return {'const': [self.value]}

    def __repr__(self):
        return 'Num(value: {})'.format(self.value)


class Op(AST):
    def __init__(self, op, children):
        self.token = self.op = op
        self._children = children

        self._replace = {
            '&&': 'and',
            '||': 'or',
            '==': 'eq',
            '!=': 'ne',
            '<=': 'le',
            '>=': 'ge',
            '<': 'lt',
            '>': 'gt',
        }

    def __repr__(self):
        return 'Op(op: {}, children: {})'.format(self.op, self._children)

    def visit(self):
        return {self._replace.get(self.op.value, self.op.value): [c.visit() for c in self._children]}


class Func(AST):
    def __init__(self, name, args):
        self.name = name
        self.args = args

    def __repr__(self):
        return 'Func(name: {}, args: {})'.format(self.name, self.args)

    def visit(self):
        return {self.name: [v.value for v in self.args]}


class Parser(object):
    def __init__(self, lexer):
        self.lexer = lexer
        self.current_token = next(self.lexer)
        log.debug('token: %s', self.current_token)

    @staticmethod
    def error():
        raise Exception('Invalid syntax')

    def eat(self, token_type):
        if self.current_token.type == token_type:
            self.current_token = next(self.lexer)
            log.debug('token: %s', self.current_token)
            return self.current_token
        else:
            self.error()

    def term(self):
        node = self.factor()

        while True:
            token = self.current_token
            if token.type == EQ:
                self.eat(EQ)
            elif token.type == GE:
                self.eat(GE)
            else:
                break
            node = Op(op=token, children=[node, self.factor()])
        return node

    def factor(self):
        token = self.current_token
        if token.type == FUNCTION:
            return self.func(token)
        elif token.type == NUMBER:
            self.eat(NUMBER)
            return Const(token)
        elif token.type == LITERAL:
            self.eat(LITERAL)
            return Const(token)
        elif token.type == LPAREN:
            self.eat(LPAREN)
            node = self.expr()
            self.eat(RPAREN)
            return node

    def func(self, name):
        self.eat(FUNCTION)
        self.eat(LPAREN)
        args = []
        while True:
            if self.current_token.type == RPAREN:
                self.eat(RPAREN)
                break
            args.append(self.expr())
            if self.current_token.type == COMMA:
                self.eat(COMMA)
            else:
                self.eat(RPAREN)
                break
        return Op(name, args)

    def expr(self):
        """
        expr    ::= term ((AND | OR) term)*
        term    ::= factor ((EQ | NE) factor)*
        factor  ::= func | const | LPAREN expr RPAREN
        func    ::= lit LPAREN expr (,expr)* RPAREN
        lit     ::= alphanum
        """
        node = self.term()

        while True:
            token = self.current_token
            if token.type == AND:
                self.eat(AND)
            elif token.type == OR:
                self.eat(OR)
            else:
                break
            node = Op(op=token, children=[node, self.term()])
        return node

    def parse(self):
        return self.expr()


def tokenize(query):
    query = query.replace(' ', '')

    idx = 0
    literal = ''
    while idx < len(query):
        for ty, op in OPERATORS:
            if query[idx: idx + len(op)] == op:
                if len(literal) > 0:
                    yield LiteralToken(literal)
                    literal = ''
                idx += len(op)
                yield Token(ty, op)
                break
        else:
            ch = query[idx]
            literal += ch
            idx += 1
    if len(literal) > 0:
        yield LiteralToken(literal)
    yield Token(EOF, None)


def compile_query(query):
    log.debug('tokens: %s', list(tokenize(query)))
    expr = Parser(tokenize(query)).expr()
    tree = expr.visit()
    log.debug('AST: %s', json.dumps(tree, indent=4))
    return tree



import json

from tornado import gen
from tornado import httputil

from cocaine.exceptions import ChokeEvent
from cocaine.services import EmptyResponse

from cocaine.proxy.helpers import fill_response_in
from cocaine.proxy.plugin import IPlugin


REQUEST_FIELDS = ('jsonrpc', 'method', 'params', 'id')


class JSONRPC(IPlugin):
    PRIMITIVE = {0: ['value', {}], 1: ['error', {}]}
    STREAMING = {0: ['write', None], 1: ['error', {}], 2: ['close', {}]}

    def __init__(self, proxy, _):
        super(JSONRPC, self).__init__(proxy)
        self._protocols = [
            (lambda tx, rx: rx == {}, self._handle_mute),
            (lambda tx, rx: rx == self.PRIMITIVE, self._handle_primitive),
            (lambda tx, rx: tx == rx == self.STREAMING, self._handle_streaming),
        ]

    @staticmethod
    def name():
        return "jsonrpc"

    def match(self, request):
        return "X-Cocaine-JSON-RPC" in request.headers

    @gen.coroutine
    def process(self, request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            JSONRPC._send_400_error(request, -32700, 'Parse error: Invalid JSON was received by the server.')
            return

        if not all(k in payload for k in REQUEST_FIELDS):
            JSONRPC._send_400_error(request, -32600, 'The JSON sent is not a valid Request object.')
            return

        name, method = payload['method'].split('.', 2)
        args = payload['params']
        chunks = payload.get('chunks', [])
        headers = {}
        if 'authorization' in request.headers:
            headers['authorization'] = request.headers['authorization']

        try:
            service = yield self.proxy.get_service(name, request)
            if service is None:
                JSONRPC._send_500_error(request, payload, 'Service not found.')
                return
        except Exception as err:
            JSONRPC._send_500_error(request, payload, err)
            return

        api = dict((data[0], data) for data in service.api.itervalues())
        if method not in api:
            JSONRPC._send_400_error(request, -32601, 'Method not found.')
            return

        name, tx_tree, rx_tree = api[method]

        try:
            for match, handle in self._protocols:
                if match(tx_tree, rx_tree):
                    result = yield handle(service, method, args, chunks, **headers)
                    break
            else:
                JSONRPC._send_400_error(request, -32000, 'Protocol type is not supported.')
                return
        except Exception as err:
            JSONRPC._send_500_error(request, payload, err)
            return

        headers = httputil.HTTPHeaders({
            'Content-Type': 'application/json-rpc'
        })
        body = {
            'jsonrpc': '2.0',
            'result': result,
            'id': payload['id'],
        }
        fill_response_in(request, 200, 'OK', json.dumps(body), headers)

    @gen.coroutine
    def _handle_mute(self, service, method, args, _, **headers):
        yield getattr(service, method)(*args, **headers)
        return

    @gen.coroutine
    def _handle_primitive(self, service, method, args, _, **headers):
        channel = yield getattr(service, method)(*args, **headers)
        result = yield channel.rx.get()
        raise gen.Return(result)

    @gen.coroutine
    def _handle_streaming(self, service, method, args, chunks, **headers):
        channel = yield getattr(service, method)(*args, **headers)
        for name, data in chunks:
            getattr(channel.tx, name)(*data)

        # TODO: Chunked.
        result = []
        try:
            while True:
                chunk = yield channel.rx.get()

                if not isinstance(chunk, EmptyResponse):
                    result.append(chunk)
        except ChokeEvent:
            pass
        finally:
            if hasattr(channel.tx, 'close'):
                channel.tx.close()

        raise gen.Return(result)

    @staticmethod
    def _send_400_error(request, code, message):
        headers = httputil.HTTPHeaders({'Content-Type': 'application/json-rpc'})
        body = {'code': code, 'message': message}
        fill_response_in(request, 400, 'Bad JSON-RPC request', json.dumps(body), headers)

    @staticmethod
    def _send_500_error(request, payload, err):
        headers = httputil.HTTPHeaders({'Content-Type': 'application/json-rpc'})
        body = {
            'jsonrpc': '2.0',
            'error': str(err),
            'id': payload['id'],
        }
        fill_response_in(request, 500, 'Internal Server Error', json.dumps(body), headers)



import logging

from tornado import web


class PingHandler(web.RequestHandler):  # pylint: disable=W0223
    def get(self):
        self.write("OK")


class LogLevel(web.RequestHandler):  # pylint: disable=W0223
    def get(self):
        lvl = self.application.logger.getEffectiveLevel()
        self.write(logging.getLevelName(lvl))

    def post(self):
        lvlname = self.get_argument("level")
        lvl = getattr(logging, lvlname.upper(), None)
        if lvl is None:
            self.write("No such level %s" % lvlname)
            return

        for name in ("cocaine.proxy.general", "cocaine.proxy.access", "cocaine.baseservice"):
            logging.getLogger(name).setLevel(lvl)
        self.write("level %s has been set" % logging.getLevelName(lvl))


class InfoHandler(web.RequestHandler):  # pylint: disable=W0223
    def get(self):
        info = self.application.proxy.info()
        self.write(info)


class UtilServer(web.Application):  # pylint: disable=W0223
    def __init__(self, proxy):
        self.proxy = proxy
        self.logger = logging.getLogger("proxy.utilserver")
        handlers = [
            (r"/ping", PingHandler),
            (r"/info", InfoHandler),
            (r"/logger", LogLevel),
        ]
        super(UtilServer, self).__init__(handlers=handlers)

    def log_request(self, handler):
        request_time = 1000.0 * handler.request.request_time()
        self.logger.info("%d %s %.2fms", handler.get_status(),
                         handler._request_summary(), request_time)
